/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*

 Maurice Clerc <Maurice.Clerc@WriteMe.com>

 Last update:
 2013-01-07 Option to force 0 and 1
 2012-01-06 Initial version

 */

#include <stdio.h>
#include "math.h"
#include <gsl/gsl_qrng.h> // Do not forget to link to gsl and gslcblas
#define halton gsl_qrng_halton
#define sobol gsl_qrng_sobol

#define DMax 10
#define nMax 5000

#define ERROR(a) {printf("\nERROR: %s\n", a);exit(-1);}

struct vector 
{  
	int size;  
	double v[DMax]; 
};

struct vectorList
{
	int size;
	struct vector V[nMax];

};

double alea (double a, double b) ;
void aleaIndex(int index[], int S);
int alea_integer (int a, int b);
double alea_normal (double mean, double std_dev);
struct vectorList quasiRand(int D,int nRand, int option,double delta);

int main (void)
{
	int d;
	int D; // Dimension
	double delta;
	int force01;
	int i;
	int nRand; // Number of quasi-random points
	int option;
	int r0, r1;
	struct vectorList qRand; // Generated list of quasi-random coordinates
	int seq,seqMax; 

	FILE * f_rand;
	f_rand=fopen ("Ltest.txt", "w");  

	D=2; // Dimension
	nRand=13; // Number of points on each dimension
	seqMax=1; // Number of consecutive sequences
	force01=1; // 1 => 0 and 1 must be in the sequence
			// 0 else
	option=1; // 0 => Halton
	// 1 => Sobol
	// 2 => 1D
	// delta is useful only for option 2
	//delta=//1./(3*(nRand+1)); // b
	//delta= 0.5/(nRand+1); // a

	fprintf(f_rand,"%i %i \n",D,nRand);
	
	if(force01==1) nRand=nRand-2;
	delta= 1./(nRand+1);

	for(seq=0;seq<seqMax;seq++)
	{
		qRand=quasiRand(D,nRand,option,delta);
//for(i=0;i<nRand;i++)
//{printf("\n");  for(d=0;d<D;d++) printf ("%.15f ", qRand.V[i].v[d]);}
		
		if(force01==1)
		{ // Insert 0 
			for(d=0;d<D;d++) 
			{
				r0=alea_integer(0,nRand-1);
//printf("\n r0 %i",r0);
				for(i=nRand-1;i>=r0;i=i-1) qRand.V[i+1].v[d]=qRand.V[i].v[d];
				qRand.V[r0].v[d]=0; 
				
			// Insert 1			
				r1: r1=alea_integer(0,nRand); if(r1==r0) goto r1;
//printf("\n r1 %i",r1);
				for(i=nRand;i>=r1;i=i-1) qRand.V[i+1].v[d]=qRand.V[i].v[d];
				qRand.V[r1].v[d]=1;				
			}

			nRand=nRand+2;
		}

		for(i=0;i<nRand;i++)
		{
			// Display
			//printf("\n"); for(d=0;d<D;d++) printf ("%.15f ", qRand.V[i].v[d]);

			// Save
			//fprintf(f_rand,"%i ",i+1);
			for(d=0;d<D;d++) fprintf (f_rand,"%.20f ", qRand.V[i].v[d] );

			fprintf(f_rand,"\n");
		}

	}
	printf("\n The resulting list is in Ltest.txt");
	return 0;
}
//==========================================================================    
struct vectorList quasiRand(int D, int nRand, int option, double delta)
{   
	/*
	 Generate nRand vectors of size D that are the coordinates of
	 nRand quasi-random points in [0,1]^D

		 For C under Linux, you need to use the GSL library:
#include <gsl/gsl_qrng.h> // Do not forget to link to gsl and gslcblas

		 */
	double deltad;
	int i;	
	int j;
	int index[nMax];
	double pi=acos(-1);
	void *qrng_q; 
	double r[nMax];
	double rTot;
	double z;

	struct vectorList qRand;

	switch(option)
	{
		default: // Halton
			case 0:
			case 1: // Sobol
				switch(option)
		{
			case 1:
				if(D>40) 
			{
				printf("\nThe embedded Sobol sequences generator can not be used for dimensions greater than 40");
				printf("\nYou should use Halton sequences (for dimensions up to 1229)");
				ERROR("\n I stop here");

			}
				qrng_q=gsl_qrng_alloc (sobol, D);

				break;

			default: 
			case 0:
				if(D>1229) 
			{printf("\nThe embedded Halton sequences generator can not be used for dimensions greater than 1229");
				printf("\nSorry");
				ERROR("\n I stop here");
			}
				qrng_q=gsl_qrng_alloc (halton, D);
		}
			gsl_qrng * q = qrng_q;	
			for (i = 0; i < nRand; i++) gsl_qrng_get (q, qRand.V[i].v);       
			gsl_qrng_free (q);
			break;

			case 2: // For dimension 1 only
				// Almost regular distribution
/*
			 // a,b
				 for (i = 0; i < nRand; i++)
				 {
					 r[i]=(double)i/(nRand+1)+ alea(-delta,delta);
	}
	r[nRand-1]=alea(1-delta,1); 
*/	
				/*
				 //c

				 for (i = 0; i <= nRand; i++)
				 {
					 r[i]=(double)i/(nRand+1)+ alea(0,delta,0);
	}
	*/	
	/*
				//d

				deltad=1./nRand;
			for (i = 0; i < nRand; i++)
		{
			r[i]=(double)i/nRand+ alea(0,deltad,0);
		}
*/
			//r[0]=alea(0,0.01,0); 
			/*
			 //e

			 deltad=1./nRand;
			 for (i = 1; i < nRand; i++)
			 {
				 r[i]=(double)i/nRand+ alea(0,deltad,0);
	}
	r[0]=alea(0,1,0);	
	*/		

			/*		
			 //f
			 for(j=0;j<8;j++)
			 for (i = 0; i < 16; i++) r[i+16*j]=(double)i/15;
			 */	
	/*		 
			 //g
			for (i = 0; i < nRand; i++)
		{
			r[i]=(double)(i+1)/(nRand+1);
		}		
*/

//h
/*	
		rTot=0;

			for (i = 0; i < nRand; i++)
		{

			r[i]=fabs(alea_normal(0.1,1));
			rTot=rTot+r[i];
		}
		// Normalise
					for (i = 0; i < nRand; i++) r[i]=r[i]/rTot;
					
		// Inverse
		for (i = 0; i < nRand; i++) r[i]=1-r[i];
*/	

//i

			for (i = 0; i < nRand; i++) r[i]=alea(0,1);
	
/*
//j. Linearly decreasing
					for (i = 0; i < nRand; i++)
					{
					r[i]=alea(0,1)+alea(0,1)-1;
					if(r[i]<0) r[i]=-r[i];
					}	
*/
/*
//k
for (i = 0; i < nRand/5; i++) r[i]=alea(0,0.01);
for (j =i; j < nRand; i++) r[i]=alea(0,1);
*/
//l
/*
deltad=pi/11; // 11
z=deltad;
for(i=0;i<nRand;i++)
{
r[i]=fabs(sin(z));
z=z+deltad;
}
for(i=0;i<nRand;i++) qRand.V[i].v[0]=r[i];
return qRand;
*/
			//-------------------------
			// Random permutation
			aleaIndex(index, nRand);

			for (i = 0; i < nRand; i++) qRand.V[i].v[0]=r[index[i]];
			break;


	}

	return qRand;
}
//==============================================================================
double alea (double a, double b) 
{				// random number (uniform distribution) in  [a b]
	double r;
	r=a+(double)rand()*(b-a)/RAND_MAX;

	return r; 
}
//==============================================================================
void aleaIndex(int index[], int S)
{
	int indexTemp[nMax];
	int length;
	int rank;
	int s;
	int t;

	length=S;
	for (s=0;s<S;s++) indexTemp[s]=s; //=index[s];

	for (s=0;s<S;s++)
	{	
		rank=alea_integer(0,length-1);
		index[s]=indexTemp[rank];
		if (rank<length-1)	// Compact
		{	
			for (t=rank;t<length-1;t++)
				indexTemp[t]=indexTemp[t+1];
		}					
		length=length-1;
	}
}
// =============================================================================
int alea_integer (int a, int b) 
{				// Integer random number in [a b]
	int ir;
	double r;

	if(a==b) return a;

	r=alea((double)a,(double)b+1); ir=floor(r);
	if(ir>b) ir=b;
	return ir;  
}
// ===========================================================
double alea_normal (double mean, double std_dev) 
{ 
	// Use the polar form of the Box-Muller transformation to obtain a pseudo
	// random number from a Gaussian distribution 

	double x1, x2, w, y1;  

	do  
	{
		x1 = 2.0 * alea (0, 1) - 1.0;
		x2 = 2.0 * alea (0, 1) - 1.0;
		w = x1 * x1 + x2 * x2;   
	}
	while (w >= 1.0); // WARNING. This may _never_ happen!

	w = sqrt (-2.0 * log (w) / w);
	y1 = x1 * w;

	if(alea(0,1)<0.5) y1=-y1; 
	y1 = y1 * std_dev + mean;
	return y1;  
}